from staze.core.error.error import Error


class OrmNotFoundError(Error):
    def __init__(
            self,
            orm_name: str,
            message: str = '',
            status_code: int = 404,
            **used_parameters) -> None:
        super().__init__(message, status_code)
        if message == '':
            self.message = \
                f"No orm {orm_name} with such" \
                f" parameters: {used_parameters}"



from enum import Enum
from staze.core.model.model import Model


class RunAppModeEnum(Enum):
    DEV = "dev"
    PROD = "prod"
    TEST = "test"


class DatabaseAppModeEnum(Enum):
    INIT = "init"
    MIGRATE = "migrate"
    UPGRADE = "upgrade"


class HelperAppModeEnum(Enum):
    SHELL = 'shell'
    DEPLOY = 'deploy'
    VERSION = 'version'

    # Custom cmds
    CMD = 'cmd'


AppModeEnumUnion = DatabaseAppModeEnum | RunAppModeEnum | HelperAppModeEnum



import os
import sys
import code
import secrets
from typing import Callable, Type

from flask_cors import CORS
from turbo_flask import Turbo
from flask_session import Session
from flask.testing import FlaskClient
from staze.core import validation
from staze.core.app.app_mode_enum import (
    RunAppModeEnum, HelperAppModeEnum, DatabaseAppModeEnum, AppModeEnumUnion)
from staze.core.log import log
from flask import Flask
from flask import cli as flask_cli
from flask.ctx import AppContext, RequestContext
from warepy import get_enum_values
from staze.core.log import log

from staze.core.service.service import Service
from staze.core.view.view import View
from .turbo_action_enum import TurboActionEnum


class App(Service):
    """Core app's class.
    """
    def __init__(
            self, 
            config: dict,
            mode_enum: AppModeEnumUnion,
            host: str,
            port: int,
            ctx_processor_func: Callable | None = None,
            each_request_func: Callable | None = None,
            first_request_func: Callable | None = None) -> None:
        # Templates by default searched within src/app, which allows to
        # integrate them directly to their logical component's folders
        self.DEFAULT_TEMPLATE_PATH = os.path.join(
            config['root_dir'], 'src/app')
        self.DEFAULT_STATIC_PATH = os.path.join(
            config['root_dir'], 'src/assets')
        self.DEFAULT_INSTANCE_PATH = os.path.join(config['root_dir'], 'var')

        # Convert all keys from given config to upper case
        self.config = self._make_upper_keys(config)
        self._assign_defaults_to_config(self.config)
        self._validate_config(self.config)

        super().__init__(self.config)
        self._mode_enum: AppModeEnumUnion = mode_enum
        self._root_dir: str = self.config['ROOT_DIR']

        self.host = host
        self.port = port
        self.native_app = self._spawn_native_app(self.config)
        self.native_app.config.from_mapping(self.config)
        self._enable_cors(self.config)
        self._enable_testing_config(self.config)
        self._add_random_hex_token_to_config()
        # Initialize turbo.js
        # https://blog.miguelgrinberg.com/post/dynamically-update-your-flask-web-pages-using-turbo-flask
        self.turbo = Turbo(self.native_app)
        # Initialize flask-session after all settings are applied
        self.flask_session = Session(self.native_app)
        self._flush_redis_session_database()

        self._init_app_daemons(
            ctx_processor_func=ctx_processor_func,
            each_request_func=each_request_func,
            first_request_func=first_request_func
        )

    @property
    def mode(self) -> str:
        return self._mode_enum.value

    @property
    def root_dir(self) -> str:
        return self._root_dir

    @property
    def test_client(self) -> FlaskClient:
        return self.native_app.test_client()

    def _assign_defaults_to_config(self, config: dict) -> None:
        if 'TEMPLATE_PATH' not in config:
            config['TEMPLATE_PATH'] = self.DEFAULT_TEMPLATE_PATH
        if 'STATIC_PATH' not in config:
            config['STATIC_PATH'] = self.DEFAULT_STATIC_PATH
        if 'INSTANCE_PATH' not in config:
            config['INSTANCE_PATH'] = self.DEFAULT_INSTANCE_PATH

    def _validate_config(self, config: dict) -> None:
        """Check if all keys in the config are correct."""
        # TODO
        pass

    def _make_upper_keys(self, mapping: dict) -> dict:
        rmap = {}
        for k, v in mapping.items():
            rmap[k.upper()] = v
        return rmap
    
    def _enable_cors(self, config: dict) -> None:
        # Enable CORS for the app's resources.
        is_cors_enabled = config.get("IS_CORS_ENABLED", None)
        if is_cors_enabled:
            CORS(self.native_app)

    def _flush_redis_session_database(self) -> None:
        # Flush redis session database if mode is not `prod`. 
        if self._mode_enum is not RunAppModeEnum.PROD: 
            if self.native_app.config.get("SESSION_TYPE", None):
                if self.native_app.config["SESSION_TYPE"] == "redis":
                    log.info(
                        "Flush session redis Database because"
                        " of non-production run.")
                    session_interface = self.native_app.session_interface
                    redis = session_interface.redis  # type: ignore
                    redis.flushdatabase()
            else:
                # Apply default null interface, basically do nothing.
                pass
    
    def _add_random_hex_token_to_config(self) -> None:
        # Generate random hex token for App's secret key, if not given in
        # config.
        if self.native_app.config["SECRET_KEY"] is None:
            self.native_app.config["SECRET_KEY"] = secrets.token_hex(16)
            
    def _enable_testing_config(self, config: dict) -> None:
        # Enable testing if appropriate mode has been set. Do not rely on enum
        # if given config explicitly sets TESTING.
        if config.get("TESTING", None) is None:
            if self._mode_enum is RunAppModeEnum.TEST:
                self.native_app.config["TESTING"] = True
            else:
                self.native_app.config["TESTING"] = False
        else:
            self.native_app.config["TESTING"] = config["TESTING"]

    def _spawn_native_app(self, config: dict) -> Flask:
        self.instance_path = self.config.get("INSTANCE_PATH")
        self.template_folder = self.config.get("TEMPLATE_PATH") 
        self.static_folder = self.config.get("STATIC_PATH")

        # Set Flask environs according to given mode.
        # Setting exactly through environs instead of config recommended by
        # Flask creators.
        # https://flask.palletsprojects.com/en/2.0.x/config/#:~:text=Using%20the%20environment,a%20previous%20value.
        if self._mode_enum not in [RunAppModeEnum.DEV, RunAppModeEnum.TEST]:
            os.environ["FLASK_ENV"] = "production"
        else:
            os.environ["FLASK_ENV"] = "development"

        return Flask(
            __name__, 
            instance_path=self.instance_path, 
            template_folder=self.template_folder, 
            static_folder=self.static_folder
        )

    def get_secret_key(self) -> str:
        """Return secret key defined in App's config."""
        return self.native_app.config["SECRET_KEY"]

    def get_url(self) -> str:
        """Return app's url.

        Hostname may be e.g. `http://127.0.0.1:5000/`.

        Warning:
            Has url resolving issues described in issue #6.
        """
        return "http://" + self.host + ":" + str(self.port) + "/"

    def get_native_app(self) -> Flask:
        """Return native app."""
        return self.native_app

    def get_instance_path(self) -> str:
        """Return app's instance path."""
        return self.native_app.instance_path

    def run(self) -> None:
        """Run Flask app."""
        self.native_app.run(
            host=self.host, port=self.port)

    def app_context(self) -> AppContext:
        return self.native_app.app_context()

    def test_request_context(self) -> RequestContext:
        return self.native_app.test_request_context()

    def register_view(self, view_class: type[View]) -> None:
        """Register given view class for the app."""
        endpoint: str
        methods: list[str]

        if view_class.ENDPOINT:
            endpoint = view_class.ENDPOINT
        else:
            endpoint = view_class.get_transformed_route()

        methods = view_class.METHODS

        validation.validate(endpoint, str)

        view_func = view_class.as_view(endpoint)

        self.native_app.add_url_rule(
            view_class.ROUTE, view_func=view_func,
            methods=methods)

    def register_error(
            self,
            error_class: type[Exception],
            handler_function: Callable) -> None:
        self.native_app.register_error_handler(error_class, handler_function)

    def push_turbo(self, action: TurboActionEnum, target: str, template_path: str, ctx_data: dict = {}) -> None:
        """Push turbo action to target with rendered from path template
        contextualized with given data.
        
        Args:
            action:
                Turbo-Flask action to perform.
            target:
                Id of HTML element to push action to.
            template_path:
                Path to template to render.
            ctx_data (optional):
                Context data to push to rendered template.
                Defaults to empty dict.
        """
        with self.native_app.app_context():
            action = action.value
            target = target
            template_path = template_path
            ctx_data = ctx_data
            exec(
                f"self.turbo.push(self.turbo.{action}"
                "(render_template("
                f"'{template_path}', **{ctx_data}), '{target}'))")

    def postbuild(self) -> None:
        """Abstract method to perform post-injection operation related to app. 
        
        Called by Assembler.

        WARNING:
            This method not working as intended now (runs not actually after
            app starting), so better to not use it temporary.
        
        Raise:
            NotImplementedError:
                If not re-implemented in children.
        """
        raise NotImplementedError(
            "Method `postbuild` hasn't been reimplemented.")

    def register_cli_cmd(self, *cmds: Callable) -> None:
        """Register cli cmds for the app."""
        # TODO: Implement.
        raise NotImplementedError("CLI cmds support in development.")
        # for cmd in cmds:
        #     self.native_app.cli.add_command(cmd)

    def register_shell_processor(self, *shell_processors: Callable) -> None:
        """Register shell processors for the app."""
        for processor in shell_processors:
            self.native_app.shell_context_processor(processor)

    def run_shell(self) -> None:
        """Adapted version of ``Flask.cli.shell_command`` by Pallets.

        Run an interactive Python shell in the context of a given
        Flask application.  The application will populate the default
        namespace of this shell according to its configuration.

        This is useful for executing small snippets of management code
        without having to manually configure the application.
        """
        banner = (
            f"Python {sys.version} on {sys.platform}\n"
            f"App: {self.native_app.import_name} [{self.native_app.env}]\n"
            f"Instance: {self.native_app.instance_path}"
        )
        ctx: dict = {}

        # Support the regular Python interpreter startup script if someone
        # is using it.
        startup = os.environ.get("PYTHONSTARTUP")
        if startup and os.path.isfile(startup):
            with open(startup) as f:
                eval(compile(f.read(), startup, "exec"), ctx)

        ctx.update(self.native_app.make_shell_context())

        # Site, customize, or startup script can set a hook to call when
        # entering interactive mode. The default one sets up readline with
        # tab and history completion.
        interactive_hook = getattr(sys, "__interactivehook__", None)

        if interactive_hook is not None:
            try:
                import readline
                from rlcompleter import Completer
            except ImportError:
                pass
            else:
                # rlcompleter uses __main__.__dict__ by default, which is
                # flask.__main__. Use the shell context instead.
                readline.set_completer(Completer(ctx).complete)

            interactive_hook()

        code.interact(banner=banner, local=ctx)

    def _init_app_daemons(
            self,
            ctx_processor_func: Callable | None,
            each_request_func: Callable | None,
            first_request_func: Callable | None) -> None:
        """Binds various background processes to the app."""
        flask_app = self.get_native_app()

        if ctx_processor_func:
            @flask_app.context_processor
            def invoke_ctx_processor_func():
                return ctx_processor_func()

        if each_request_func:
            @flask_app.before_request
            def invoke_each_request_func():
                return each_request_func()

        if first_request_func:
            @flask_app.before_first_request
            def invoke_first_request_func():
                return first_request_func()



"""Module for resolving metaclasses conflicts.

Refs:
    http://www.phyast.pitt.edu/~micheles/python/metatype.html
"""

metadic={}

def _generatemetaclass(bases,metas,priority):
    trivial=lambda m: sum([issubclass(M,m) for M in metas],m is type)
    # hackish!! m is trivial if it is 'type' or, in the case explicit
    # metaclasses are given, if it is a superclass of at least one of them
    metabs=tuple([mb for mb in map(type,bases) if not trivial(mb)])
    metabases=(metabs+metas, metas+metabs)[priority]
    if metabases in metadic: # already generated metaclass
        return metadic[metabases]
    elif not metabases: # trivial metabase
        meta=type
    elif len(metabases)==1: # single metabase
        meta=metabases[0]
    else: # multiple metabases
        metaname="_"+''.join([m.__name__ for m in metabases])
        meta=makecls()(metaname,metabases,{})
    return metadic.setdefault(metabases,meta)

def makecls(*metas,**options):
    """Class factory avoiding metatype conflicts. The invocation syntax is
    makecls(M1,M2,..,priority=1)(name,bases,dic). If the base classes have
    metaclasses conflicting within themselves or with the given metaclasses,
    it automatically generates a compatible metaclass and instantiate it.
    If priority is True, the given metaclasses have priority over the
    bases' metaclasses"""

    priority=options.get('priority',False) # default, no priority
    return lambda n,b,d: _generatemetaclass(b,metas,priority)(n,b,d)


import os
from typing import Callable
from staze.core.app.app_mode_enum import AppModeEnumUnion
from staze.core.app.app import App
from staze.core.assembler.assembler import Assembler
from staze.core.socket.default_sock_error_handler import (
    default_sock_error_handler)
from staze.core.service.service import Service
from staze.core.view.view import View
from staze.core.error.error import Error
from staze.core.socket.sock import Sock


class Build:
    """Proxy mapping class with collection of initial project instances to be
    builded by assembler.

    Args:
        version (optional):
            Version of the project. Defaults to '0.0.0'
        env_path (optional):
            Path to file with environment variables. Defaults to '.env'
        config_dir (optional):
            Directory of configuration files. Defaults to './src/configs'
        service_classes (optional):
            List of Service classes to be initialized. Defaults to empty list
        view_classes (optional):
            List of View classes to be initialized. Defaults to empty list
        sock_classes (optional):
            List of Sock classes to be initialized. Defaults to empty list
        shell_processors (optional):
            List of Callables that will serve shell processor function. Note
            that shell processor callable should return dictionary with
            namespace for the shell.
            Defaults to empty list
        default_error_handler (optional):
            Callable to be called on Error reaching top level of program.
            Defaults to None, e.g. app's default logic will be used
        default_builtin_error_handler (optional):
            Callable to be called on Exception reaching top level of program.
            Defaults to None, e.g. app's default logic will be used
        ctx_processor_func (optional):
            Callable to be called on Flask.context_processor.
            Defaults to None
        each_request_func (optional):
            Callable to be called on Flask.before_request.
            Defaults to None
        first_request_func (optional):
            Callable to be called on Flask.before_first_request.
            Defaults to None
    """
    def __init__(
            self,
            *,
            version: str = '0.0.0',
            env_path: str = '.env',
            config_dir: str = './src/configs',
            service_classes: list[type[Service]] = [],
            view_classes: list[type[View]] = [],
            shell_processors: list[Callable] = [],
            sock_classes: list[type[Sock]] = [],
            default_error_handler: Callable | None = None,
            default_builtin_error_handler: Callable | None = \
                None,
            ctx_processor_func: Callable | None = None,
            each_request_func: Callable | None = None,
            first_request_func: Callable | None = None) -> None:
        self.version = version
        self.env_path = env_path
        self.config_dir = config_dir
        self.service_classes = service_classes
        self.view_classes = view_classes
        self.shell_processors = shell_processors
        self.ctx_processor_func = ctx_processor_func
        self.each_request_func = each_request_func
        self.first_request_func = first_request_func
        self.sock_classes = sock_classes
        self.default_error_handler = default_error_handler
        self.default_builtin_error_handler = default_builtin_error_handler

    def build_app(
            self,
            mode_enum: AppModeEnumUnion,
            host: str = 'localhost',
            port: int = 5000,
            root_dir: str = os.getcwd()) -> App:
        """Allows manually build assembler without immediate app running."""
        assembler = Assembler(
            mode_enum=mode_enum,
            host=host,
            port=port,
            root_dir=root_dir,
            build=self)
        return assembler.app



from staze.core.error.error import Error
from staze.core.validation import validate


class ParsingError(Error):
    pass


class IntParsingError(ParsingError):
    def __init__(
            self,
            parsed_str: str,
            message: str | None = None,
            status_code: int | None = None) -> None:
        super().__init__(message, status_code)

        validate(parsed_str, str, 'Parsed str')
        if message is None:
            self.message = f'{parsed_str} is not parseable to int'
        

class KeyParsingError(ParsingError):
    def __init__(
            self,
            parsed_map: dict,
            failed_key: str,
            message: str | None = None,
            status_code: int | None = None) -> None:
        super().__init__(message, status_code)

        validate(parsed_map, dict, 'Parsed map')
        validate(failed_key, str, 'Failed key')

        if message is None:
            self.message = f'{parsed_map} has no key: \'{failed_key}\''



# NOTE: It's extremely important to not set return typehint in decorators with wraps,
# if you want to save your wrapped function's docstring (occurred in VsCode's Pylance Python Language Server).
# BUT, you should set return type Callable in decorators with arguments, or interpreters like PyRights will bind
# decorated functions to Unknown return type.
from functools import wraps
from typing import Callable

from warepy import format_message
from staze.core.log import log
from flask import session, redirect, url_for


def login_required(
    endpoint_if_not_logged: str, 
    allowed_types: list[str] | None = None, 
    endpoint_if_not_allowed: str | None = None
) -> Callable:
    """Check if user logged in before giving access to wrapped view.
    
    If user is not logged in, redirect him to the login page.
    If user doesn't have access to the view (i.e. his type is not in `allowed_types`), redirect him to backup page.

    Login checked against `flask.session` object with key specified under argument `user_id_session_key`.
    ```py
        session = {
            "user": {
                "username": USERNAME,  # To display during errors, etc.
                "type": USER_TYPE,  # User type to check against argument `allowed_types`.
                # ... another restriction-free fields.
            }
            # ... another session fields.
        }
    ```

    Args:
        endpoint_if_not_logged: 
            Endpoint to redirect to if user is not logged in.
        allowed_types: 
            Types of users that should have access to the view. Defaults to None, i.e. all logged users have access.
        endpoint_if_not_allowed: 
            Endpoint to redirect to if user not in allowed types to access wrapped view. 
            Defaults to None. Should be set if `allowed_types` argument given.

    Raise:
        ValueError:
            If `allowed_types` given, but `endpoint_if_not_allowed` is not.
    """
    if allowed_types and not endpoint_if_not_allowed is None or not allowed_types and endpoint_if_not_allowed:
        raise ValueError(format_message("List of allowed types to view is provided, but endpoint of not-logged users is not."))

    def decorator(view: Callable):
        @wraps(view)
        def inner(**kwargs):
            result = None
            error_message = None

            if session.get("user", None) is None:
                error_message = format_message("Reject request of unauthorized user to view: {}", view.__name__)
                result = redirect(url_for(endpoint_if_not_logged))
            # Variable `endpoint_if_not_allowed` checked here for the second time since Pyright gives error on redirect line.
            elif allowed_types is not None and endpoint_if_not_allowed:
                if session["user"]["type"] not in allowed_types:
                    error_message = format_message("Reject request of user {} with type {} to view {}.", [session["user"]["username"], session["user"]["name"], view.__name__])
                    result = redirect(url_for(endpoint_if_not_allowed))
            
            # Check if error occured, else normally call view. Finally return result with error or view output.
            if error_message is not None:
                log.warning(error_message)
            else:
                result = view(**kwargs)
            return result
        return inner
    return decorator


from .error.error import Error


class NotFoundError(Error):
    def __init__(
            self,
            message: str = 'Requested resource is not found',
            status_code: int = 404) -> None:
        super().__init__(message, status_code)



import os

from pytest import fixture
from staze.core.assembler.assembler import Assembler
from staze.core.assembler.build import Build
from staze.core.app.app_mode_enum import RunAppModeEnum
from staze.core.app.app import App
from staze.core.database.database import Database
from staze.core.log import log
from staze.tests.blog.app.user.user_service import UserService


@fixture
def assembler_dev(
        blog_build: Build, default_host: str, default_port: int,
        blog_root_dir: str):
    return Assembler(
        build=blog_build,
        mode_enum=RunAppModeEnum.DEV,
        host=default_host,
        port=default_port,
        root_dir=os.path.join(os.getcwd(), blog_root_dir))


class TestAssembler():
    def test_build(self, assembler_dev: Assembler):
        app: App = App.instance()
        database: Database = Database.instance()
        user_service: UserService = assembler_dev.custom_services['user']



from typing import Callable
from pytest import fixture
from flask_socketio import SocketIOTestClient
from flask.testing import FlaskClient
from werkzeug.test import TestResponse

from staze.core.app.app import App
from staze.core import validation, parsing
from staze.core.database.database import Database
from staze.core.socket.socket import Socket
from staze.core.test.http_client import HttpClient


class Test:
    @fixture
    def http(self, app: App, client: FlaskClient) -> HttpClient:
        return HttpClient(client)

    @fixture
    def app(self):
        app: App = App.instance()
        yield app

    @fixture
    def db(self, app: App):
        database: Database = Database.instance()

        with app.app_context():
            database.drop_all()
            database.create_all()

        yield database

        with app.app_context():
            database.drop_all()

    @fixture
    def socket(self) -> Socket:
        return Socket.instance()

    @fixture
    def client(self, app: App) -> FlaskClient:
        return app.test_client

    @fixture
    def socket_client(self, app: App, socket: Socket) -> SocketIOTestClient:
        # https://github.com/miguelgrinberg/Flask-SocketIO/blob/main/test_socketio.py
        return socket.get_test_client()

    @fixture
    def root_dir(self, app: App) -> str:
        return app.root_dir



"""
Invoke cli interface when staze executed as a script:
```py
python -m staze
```
"""
try:
    from .core.cli import cli
except ImportError:
    from staze.core.cli import cli


if __name__ == "__main__":
    cli.main()


import re
from typing import Any
from enum import EnumMeta
from staze.core.log import log
from staze.core.validation.validation_error import (
    ValidationError, ReValidationError)


def validate(
        obj: Any, expected_type: type | list[type],
        obj_name: str = 'Entity', strict: bool = False) -> None:
    if isinstance(expected_type, type):
        if strict:
            if type(obj) is not expected_type:
                raise ValidationError(obj_name, expected_type)
        else:
            if not isinstance(obj, expected_type):
                raise ValidationError(obj_name, expected_type)
    elif type(expected_type) is list:
        found: bool = False

        for type_ in expected_type:
            if type(obj) is type_:
                found = True
        
        if not found:
            raise ValidationError(obj_name, expected_type)
    else:
        raise TypeError('Expected type should be `type` type')


def validate_re(string: str, pattern: str) -> None:
    """Validate given string using re.match.
    
    Raises:
        ValidationError:
            String does not match given pattern.
    """
    if not re.match(pattern, string):
        raise ReValidationError(validated_name=string, pattern=pattern)



from dataclasses import dataclass, fields
from typing import Any
import types
from staze.core.log import log
from typing import ClassVar
from warepy import snakefy
from pydantic import BaseModel


class Model(BaseModel):
    """Basic interface dataclass."""
    # Name to be excluded from final formatted name
    _base_name: str = 'Model'
    _formatted_name: str | None = None

    @property
    def formatted_name(self) -> str:
        return self._get_formatted_name()

    @property
    def api_dict(self) -> dict:
        return self.get_api_dict()

    def get_api_dict(self, formatted_name: str | None = None) -> dict:
        """Translates interface to dict structure for API.
        
        Args:
            formatted_name (str, optional):
                Custom formatted name as external dictionary key.
                Defaults to class's base defined (property formatted_name).

        Returns:
            dict:
                Dictionary with interface decomposed to structure for API:
        ```
        {
           model_formatted_name: {
                ...model structure    
            }
        }
        ```
        """
        _formatted_name: str

        if formatted_name:
            _formatted_name = formatted_name  
        else:
            _formatted_name = self._get_formatted_name()

        return {
            _formatted_name: self.dict()
        }

    @classmethod
    def _get_formatted_name(cls) -> str:
        name: str

        if cls._formatted_name:
            name = cls._formatted_name
        else:
            name = snakefy(cls.__name__.replace(cls._base_name, ''))

        return name



import inspect
from typing import Callable
from werkzeug.test import TestResponse
from staze.core import validation, parsing
from staze.core.log import log
from flask.testing import FlaskClient


class HttpClient():
    def __init__(self, client: FlaskClient) -> None:
        self._client: FlaskClient = client

    def get(
            self, url: str, asserted_status_code: int = 200,
            **kwargs) -> TestResponse:
        # Join method name and function to call inner resolver
        # inspect.stack() is for resolving self method name, ref:
        #   https://stackoverflow.com/a/5067654
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def post(
            self, url: str, asserted_status_code: int = 200,
            **kwargs) -> TestResponse:
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def delete(
            self, url: str, asserted_status_code: int = 200,
            **kwargs) -> TestResponse:
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def patch(
            self, url: str, asserted_status_code: int = 200,
            **kwargs) -> TestResponse:
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def put(
            self, url: str, asserted_status_code: int = 200,
            **kwargs) -> TestResponse:
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def _get_test_response(
            self,
            stack: list[inspect.FrameInfo],
            url: str,
            asserted_status_code: int,
            **kwargs) -> TestResponse:
        request: str = ' '.join([stack[0][3], url])
        return self._resolve_request(request, asserted_status_code, **kwargs)


    def _resolve_request(
            self,
            request: str,
            asserted_status_code: int,
            **request_kwargs) -> TestResponse:
        response: TestResponse
        test_client_method: Callable
        method: str
        url: str

        validation.validate(request, str)
        validation.validate(asserted_status_code, int)

        # Request example: 'get /users/1'
        method, url = request.split(' ')

        # Also can accept uppercase 'GET ...'
        method = method.lower()

        match method:
            case 'get':
                test_client_method = self._client.get
            case 'post':
                test_client_method = self._client.post
            case 'put':
                test_client_method = self._client.put
            case 'patch':
                test_client_method = self._client.patch
            case 'delete':
                test_client_method = self._client.delete
            case _:
                raise ValueError(f'Method {method} is not supported')

        response = test_client_method(url, **request_kwargs)

        assert response.status_code == asserted_status_code, response.json

        return response



import os
import sys

from pytest import fixture
from staze import Build

from staze.tests.blog.build import build as _blog_build
from staze.core.assembler.assembler_test import assembler_dev
from staze.core.log import log


@fixture
def blog_build() -> Build:
    build = _blog_build
    build.config_dir = 'configs'

    return build


@fixture
def default_host() -> str:
    return '127.0.0.1'


@fixture
def default_port() -> int:
    return 6000


@fixture
def blog_root_dir() -> str:
    return 'staze/tests/blog'



from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable

from staze.core.service.service_ie import ServiceModel

if TYPE_CHECKING:
    from staze.core.app.app import Staze
    from staze.core.hints import AppModeEnumUnion

 
class StazeServiceModel(ServiceModel):
    """Injection cell with app itself which is required in any build."""
    service_class: type[Staze]
    mode_enum: AppModeEnumUnion
    host: str
    port: int
    ctx_processor_func: Callable | None = None
    each_request_func: Callable | None = None
    first_request_func: Callable | None = None



import re

from flask.views import MethodView
from warepy import Singleton, format_message
from staze.core.log import log

from staze.core.noconflict import makecls


class View(MethodView):
    """Presents API source with HTTP analog methods to be registered in app routing.
    
    Contains general methods `get`, `post`, `put` and `delete` according to same HTTP methods 
    and should be re-implemented in children classes.
    
    Refs:
        https://flask.palletsprojects.com/en/2.0.x/views/#method-views-for-apis
    """
    __metaclass__ = makecls()

    # Route will be the same for all methods
    ROUTE: str
    # View methods. Note that this list completely overrides HttpMethodEnum
    # for this view
    METHODS: list[str] = ['GET', 'POST', 'PUT', 'PATCH', 'DELETE']
    ENDPOINT: str | None = None

    # List of decorators to apply to all view's methods.
    # decorators = [log.catch]  
    # To extend decorators in child class, use `decorators = View.decorators + [your_shiny_decorator]` 
    # in your class variable definition.

    @classmethod
    def get_transformed_route(cls) -> str:
        """Return route transformed to endpoint format.

        This method used by app to get endpoint if it's not given at assembling
        stage.
        
        Example:
        ```python
        v = ViewModel(view_class=MyViewClass, route='/my/perfect/route')
        v.get_transformed_route()
        # 'my.perfect.route'
        ```
        """
        res_route: str = ''
        route_pieces: list[str] = cls.ROUTE.split('/')

        for piece in route_pieces:
            if re.match(r'\<.+\>', piece):
                # Remove arrows around route, so /user/<id> transforms to
                # user.id
                res_route += piece[1:len(piece)-2]
            else:
                res_route += piece

        return res_route




from typing import Callable, Any, TypeVar

from werkzeug.datastructures import MultiDict

from staze.core.parsing.parsing_error import (
    IntParsingError, KeyParsingError, ParsingError)
from staze.core.query_parameter_error import QueryParameterError

from staze.core.filter_query_enum import FilterQueryEnum
from staze.core.validation import validate
from staze.core.database.database import Database


ParsedEntity = TypeVar('ParsedEntity', bound=Any)


def parse(entity: Any, expected_type: type[ParsedEntity]) -> ParsedEntity:
    """Return entity if its type matches `expected_type`.
    
    This is only strict type checking. For advanced checking consider other
    functions of this module.
    """
    if type(entity) is expected_type:
        return entity
    else:
        raise ParsingError(f'Entity {entity} should have type {expected_type}')


def parse_bool(entity: str | bool) -> bool:
    res: bool

    validate(entity, [bool, str], 'Entity')
    
    if type(entity) is str:
        match entity:
            case 'true':
                res = True
            case 'false':
                res = False
            case _:
                raise ParsingError(
                    'Str entity to transform to bool should have value'
                    ' `true` or `false`')
    elif type(entity) is bool:
        res = entity
    else:
        raise

    return res


def parse_int(entity: int | str) -> int:
    """Validate given entity and try to convert it to integer.

    If str accepted, converts it to integer.
    
    Raise:
        IntParsingError:
            Cannot parse entity

    Return:
        Entity converted to integer
    """
    res: int

    validate(entity, [int, str], 'Entity')

    if type(entity) is int:
        res = entity
    elif type(entity) is str:
        try:
            res = int(entity)
        except ValueError:
            raise IntParsingError(entity)
    else:
        raise

    return res


def parse_key(
        key: str, entity: dict,
        post_validation_type: type | list[type] | None = None,
        default: Any = None,
        strict: bool = False) -> Any:
    """Return value parsed from entity by given key.
    
    Apply validation.validate() on result if `post_validation_type` given.
    """
    validate(entity, dict, 'Map to parse')
    validate(key, str, 'Key')

    try:
        value: Any = entity[key]
    except KeyError:
        if default is None:
            raise KeyParsingError(entity, key)
        else:
            value = default
    else: 
        if post_validation_type is not None:
            validate(
                post_validation_type, [type, list], 'Post validation type',
                strict=False)
            validate(
                value, post_validation_type, key.capitalize(), strict=strict)

    return value



import os
from typing import Literal

from loguru import logger as loguru
from warepy import Singleton


class log(Singleton):
    """Logger tool responsible of writing all actions to logs.

    Simply said - it is a extra layer over `loguru.log` for keeping one log
    sink through all program.
    """
    DEFAULT_LOG_PARAMS = {
        "path": "./var/logs/system.log",
        "level": "DEBUG",
        "format":
            "{time:%Y.%m.%d at %H:%M:%S:%f%z} | {level} | {extra} >> {message}",
        "rotation": "10 MB",
        "serialize": False
    }

    native_log = loguru

    catch = native_log.catch

    debug = native_log.debug
    info = native_log.info
    warning = native_log.warning
    error = native_log.error
    critical = native_log.critical

    @classmethod
    def _create(cls, *args, **kwargs) -> int:
        return cls.native_log.add(*args, **kwargs)

    @classmethod
    def _remove(cls, id: int) -> None:
        cls.native_log.remove(id)

    @classmethod
    def get_native_log(cls):
        return cls.native_log

    @classmethod
    def bind(cls, **kwargs):
        return cls.native_log.bind(**kwargs)

    @classmethod
    def configure(
            cls,
            *, 
            path: str, 
            format: str, 
            rotation: str, 
            level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            serialize: bool,
            delete_old: bool = False) -> None:
        """Init log model instance depending on given arguments. 

        Just a helpful method to recognize which arguments should be added with
        type hinting, and which is remain static.
        
        Remove previous old log if `delete_old=True`.
        """
        # TODO: Add more clarified explanation/logic to this deletion.  It
        # should be either documented to user or refactored with security
        # considerations
        if (
                delete_old
                and os.path.isfile(path)
                # Ensure that specified file is log for security
                and path.split('.')[-1] == 'log'):
            os.remove(path)

        cls._create(
            path, 
            format=format, 
            level=level,
            compression="zip", 
            rotation=rotation, 
            serialize=serialize
        )


@log.catch
def raise_error_to_log(error: Exception) -> None:
    """Raise error with default logger catching."""
    raise error



import os
import re
import sys
import pytest
from typing import get_args

from warepy import (
    match_enum_containing_value,
    get_enum_values, get_union_enum_values
)
from staze.core.cli.cli_error import CLIError
from staze.core.log import log

from staze import __version__ as staze_version
from staze.core.assembler.assembler import Assembler
from staze.core import validation, parsing
from staze.core.app.app_mode_enum import (
    RunAppModeEnum, DatabaseAppModeEnum, HelperAppModeEnum, AppModeEnumUnion)
from staze.core.cli.cli_input import CliInput


def main() -> None:
    args: CliInput = _parse_input()

    match args.mode_enum:
        case HelperAppModeEnum.VERSION:
            print(f"Staze {staze_version}")
            exit()
        case _:
            # Create and build assembler
            assembler = Assembler(
                mode_enum=args.mode_enum,
                mode_args=args.mode_args,
                host=args.host,
                port=int(args.port))

            assembler.run()


def _parse_input() -> CliInput:
    args: list[str] = sys.argv
    check_other_args: bool = True

    model_kwargs: dict = {}
    model_kwargs['mode_args'] = []


    match args[1]:
        case 'version':
            if len(args) > 2:
                raise CLIError(
                    'Mode `version` shouldn\'t be'
                    ' followed by any other arguments')

            model_kwargs['mode_enum'] = HelperAppModeEnum.VERSION
            check_other_args = False 
        case _:
            try:
                # Find enum where mode assigned
                mode_enum_class = match_enum_containing_value(
                    args[1], *get_args(AppModeEnumUnion))
            except ValueError:
                raise CLIError(
                    f'Unrecognized mode: {args[1]}')
            else:
                # Create according enum with mode value
                model_kwargs['mode_enum'] = mode_enum_class(args[1])

    if check_other_args:
        # Traverse other args
        for ix, arg in enumerate(args[2:]):
            ix = ix + 2

            match arg:
                case '-h':
                    if not isinstance(model_kwargs['mode_enum'], RunAppModeEnum):
                        raise CLIError(
                            'Flag -h applicable only to Run modes:'
                            f' {get_enum_values(RunAppModeEnum)}')
                    elif '-h' in model_kwargs:
                        raise CLIError('Flag -h has been defined twice')
                    else:
                        try:
                            host = args[ix+1]
                        except KeyError:
                            raise CLIError(
                                'No host specified for defined flag -h')
                        else:
                            # Pattern from: https://stackoverflow.com/a/36760050
                            validation.validate_re(
                                host,
                                r'^((25[0-5]|(2[0-4]|1\d|[1-9]|)\d)(\.(?!$)|$)){4}$')
                        model_kwargs['host'] = host
                case '-p':
                    if not isinstance(model_kwargs['mode_enum'], RunAppModeEnum):
                        raise CLIError(
                            'Flag -p applicable only to Run modes:'
                            f' {get_enum_values(RunAppModeEnum)}')
                    elif '-p' in model_kwargs:
                        raise CLIError('Flag -p has been defined twice')
                    else:
                        try:
                            port = args[ix+1]
                        except KeyError:
                            raise CLIError(
                                'No port specified for defined flag -p')
                        else:
                            validation.validate_re(
                                port,
                                r'^\d+$')
                        model_kwargs['port'] = port
                case _:
                    # In all other cases, write results to mode_args as it is,
                    # this is required, e.g. in pytest as well as in all other
                    # plugins and custom commands
                    model_kwargs['mode_args'].append(arg)

    return CliInput(**model_kwargs)


if __name__ == "__main__":
    main()



from typing import TYPE_CHECKING

from flask_socketio import (
    SocketIO, send, emit, join_room, leave_room, SocketIOTestClient)

from staze.core.service.service import Service
from staze.core.log import log

if TYPE_CHECKING:
    from staze.core.app.app import App


class Socket(Service):
    def __init__(self, config: dict, app: 'App') -> None:
        super().__init__(config)
        self.app = app
        cors_allowed_origin = self.config.get('cors_allowed_origin', '*')

        log.info(f'Socket cors allowed: {cors_allowed_origin}')
        self.socketio: SocketIO = SocketIO(
            self.app.get_native_app(), cors_allowed_origins=cors_allowed_origin)

    def get_socketio(self) -> SocketIO:
        return self.socketio

    def get_test_client(self, namespace: str | None = None) -> SocketIOTestClient:
        return self.socketio.test_client(
            self.app.get_native_app(), namespace=namespace)

    def emit(
        self,
        event,
        data=None,
        room=None,
        include_self=True,
        namespace=None,
        callback=None) -> None:
        """Emit a custom event to one or more connected clients."""
        return self.socketio.emit(
            event, 
            data, 
            room=room,
            include_self=include_self,
            namespace=namespace,
            callback=callback)

    def send(
            self,
            data,
            room=None,
            include_self=True,
            namespace=None,
            callback=None,
            json=False) -> None:
        """Send a message to one or more connected clients."""
        return self.socketio.send(
            data,
            room=room,
            include_self=include_self,
            namespace=namespace,
            callback=callback,
            json=json)



from typing import Callable

from flask_socketio import SocketIO, Namespace
from staze.core.log import log

from .socket import Socket
from .default_sock_error_handler import default_sock_error_handler


class Sock(Namespace):
    NAMESPACE: str
    ERROR_HANDLER: Callable = default_sock_error_handler

    def __init__(self, namespace=None):
        super().__init__(namespace)
        self.socket: Socket = Socket.instance()

    def on_connect(self):
        self.socket.send('Connected')

    def on_disconnect(self):
        self.socket.send('Disconnected')

    def on_message(self, message):
        self.socket.send(f'Hello {message}!')



from typing import Any
from staze.core.error.error import Error


class ValidationError(Error):
    def __init__(
            self,
            validated_name: str,
            expected_type: type | list[type],
            message: str | None = None,
            status_code: int | None = None) -> None:
        super().__init__(message, status_code)

        if message is None:
            if isinstance(expected_type, type):
                self.message = \
                    f'{validated_name} should have type:' \
                    f' {expected_type.__name__}'
            elif type(expected_type) is list:
                self.message = \
                    f'{validated_name} should have one type of the following' \
                    f' list: {[type_.__name__ for type_ in expected_type]}'
            else:
                raise TypeError('Unrecognized type of `expected_type`')


# ReValidationError inherits base Error, not ValidationError, since some old
# features differ from ValidationError
class ReValidationError(Error):
    def __init__(
            self,
            validated_name: str,
            pattern: str,
            message: str | None = None,
            status_code: int | None = None) -> None:
        super().__init__(message, status_code)
        
        if message is None:
            self.message = \
                f'{validated_name} should implement pattern {pattern}'



__version__ = '0.1.0'

from .core.error.error import Error
from .core.model.model import Model
from .core.not_found_error import NotFoundError
from .core.view.view import View
from .core.assembler.build import Build
from .core.service.service import Service
from .core.app.app import App
from .core.database.database import Database
from .core.test.test import Test
from .core.test.mock import Mock
from .core.login_required_dec import login_required
from .core.log import log
from .core.database.orm_not_found_error import OrmNotFoundError
from .core.socket.socket import Socket
from .core.socket.sock import Sock
from .core.test.http_client import HttpClient
from .core.query_parameter_error import QueryParameterError
from .core.filter_query_enum import FilterQueryEnum

# Modules import
from .core import validation, parsing



from __future__ import annotations
import os
import sys
import importlib.util
from typing import TYPE_CHECKING, Any
from dotenv import load_dotenv

from warepy import (
    format_message, load_yaml, get_enum_values, Singleton
)
from flask_socketio import SocketIO
import pytest
from staze.core.assembler.assembler_error import AssemblerError
from staze.core.app.app_mode_enum import (
    DatabaseAppModeEnum, RunAppModeEnum, AppModeEnumUnion, HelperAppModeEnum)
from staze.core.error_handler import ErrorHandler

from staze.core.socket.socket import Socket
from staze.core.service.service import Service
from staze.core.log import log
from staze.core.error.error import Error
from staze.core.model.config import Config

from staze.core.database.database import Database
from staze.core.app.app import App

if TYPE_CHECKING:
    from .build import Build


class Assembler(Singleton):
    """Assembles all project instances and initializes it.

    Acts automatically and shouldn't be inherited directly by project in any
    form.

    TODO: Write all init args
    Args:
        extra_configs_by_name (optional):
            Configs to be appended to appropriate ones described in Build
            class. Defaults to None.
            Contain config name as key (which is compared to Config
            name field) and configuration mapping as value, e.g.:
    ```python
    extra_configs_by_name = {
        "app": {"TESTING": True},
        "database": {"database_uri": "sqlite3:///:memory:"}
    }
    ```
        root_dir (optional):
            Root dir to execute project from. Defaults to `os.getcwd()`.
            Required in cases of calling this function not from actual
            project root (e.g. from tests) to set root path explicitly.
    """
    def __init__(
            self, 
            mode_enum: AppModeEnumUnion,
            mode_args: list[str] = [],
            source_filename: str = 'build',
            build: Build | None = None,
            host: str = 'localhost',
            port: int = 5000,
            root_dir: str = os.getcwd(),
            extra_configs_by_name: dict[str, Any] | None = None) -> None:
        # Define attributes for getter methods to be used at builder
        # Do not set extra_configs_by_name to None at initialization, because
        # `get()` method called from this dictionary
        self.extra_configs_by_name = {}
        self.root_dir = root_dir
        self.socket_enabled: bool = False

        self.mode_enum: AppModeEnumUnion = mode_enum
        self.mode_args: list[str] = mode_args

        # Load build from module
        if build:
            self.build = build
        else:
            self.build = self._load_target_build(source_filename)

        # Environs should be loaded from app's root directory
        load_dotenv(os.path.join(self.root_dir, self.build.env_path))

        # Use build in further assignment
        self.service_classes = self.build.service_classes
        self.view_classes = self.build.view_classes
        self.shell_processors = self.build.shell_processors
        self.ctx_processor_func = self.build.ctx_processor_func
        self.each_request_func = self.build.each_request_func
        self.first_request_func = self.build.first_request_func
        self.sock_classes = self.build.sock_classes

        self.default_error_handler = self.build.default_error_handler
        self.default_builtin_error_hanlder = \
            self.build.default_builtin_error_handler

        # Load env or use build default
        self.config_dir: str
        env_config_dir = os.getenv('CONFIG_DIR')
        if isinstance(env_config_dir, str) and env_config_dir:
            log.info(
                f'Env CONFIG_DIR specified to {env_config_dir},'
                ' use it by default')
            self.config_dir = env_config_dir
        else:
            self.config_dir = self.build.config_dir

        self._build_configs(self.config_dir)
        
        self._build_builtin_services(mode_enum, host, port)
        self._build_builtin_helpers()

        # Namespace to hold all initialized services. Should be used only for
        # testing purposes, when direct import of services are unavailable
        self._custom_services: dict[str, Any] = {}

        # Add extra configs
        if extra_configs_by_name:
            self.extra_configs_by_name = extra_configs_by_name

        self._register_self_singleton()

        self._build_log()
        self._build_custom_services()
        self._build_custom_views()
        self._build_custom_shell_processors()
        self._build_custom_cli_cmds()
        self._build_custom_socks()

        # Call postponed build from created App.
        try:
            self.app.postbuild()
        except NotImplementedError:
            pass

    @property
    def custom_services(self):
        return self._custom_services

    def _register_self_singleton(self):
        """Register self instance to Singleton instance.
        
        This should be done at initialization to avoid built services Assembler
        initialization problem via `Assembler.instance()` call.
        """
        type(self.__class__).instances[self.__class__] = self

    def _load_target_build(self, source_filename: str) -> Build:
        """Load target module spec from location, where cli were called."""
        # Add root_dir to sys.path to avoid ModuleNotFoundError during lib
        # importing
        sys.path.append(self.root_dir)

        module_location = os.path.join(self.root_dir, source_filename + ".py")
        module_spec = importlib.util.spec_from_file_location(
            source_filename, module_location)
        if module_spec and module_spec.loader:
            main_module = importlib.util.module_from_spec(module_spec)
            module_spec.loader.exec_module(main_module)
            # TODO: Add validation of result build?
            return main_module.build
        else:
            raise NameError(
                format_message(
                    "Could not resolve module spec for location: {}.",
                    module_location))

    def _build_configs(self, config_dir: str) -> None:
        """Traverse through config files under given config_dir and create 
        Configs from them.

        Name taken from filename of config and should be the same as specified 
        at config's target service_class.name.

        Names can contain additional extension like `name.prod.yaml` according
        to appropriate app modes. These configs launched per each mode. Config
        without this extra extension, considered `prod` moded.
        """
        self.config_classes: list[Config] = []

        config_path: str = os.path.join(self.root_dir, config_dir)
        source_map_by_name: dict[str, dict[RunAppModeEnum, str]] = \
            Config.find_config_files(config_path)

        for name, source_map in source_map_by_name.items():
            self.config_classes.append(Config(
                name=name,
                source_by_app_mode=source_map))

    def _build_builtin_services(
            self, mode_enum: AppModeEnumUnion, host: str, port: int) -> None:
        """Assign builting service cells if configuration file for its service
        exists.
        """
        # First service to be initialized is always the App
        self.app: App = App(
            mode_enum=mode_enum, host=host, port=port, 
            config=self._assemble_service_config('app'),
            ctx_processor_func=self.ctx_processor_func,
            each_request_func=self.each_request_func,
            first_request_func=self.first_request_func)
        layers_to_log: list[str] = []

        # Enable only modules with specified configs.
        # TODO:
        #   Instead of catching ValueError, which could be anything, make
        #   own custom error inside find_by_name()
        if self.config_classes:
            try:
                Config.find_by_name('database', self.config_classes)
            except ValueError:
                pass
            else:
                self.database: Database = Database(
                    config=self._assemble_service_config('database'))

                # Perform Database postponed setup
                self._perform_database_postponed_setup()
                layers_to_log.append('database')
            try:
                Config.find_by_name('socket', self.config_classes)
            except ValueError:
                pass
            else:
                self.socket = Socket(
                    config=self._assemble_service_config('socket'),
                    app=self.app)
                self.socket_enabled = True
                layers_to_log.append('socket')
            
            if layers_to_log:
                log.info(f'Enabled layers: {", ".join(layers_to_log)}')

    def _build_builtin_helpers(self) -> None:
        self._build_error_handler()

    def run(self):
        if isinstance(self.mode_enum, RunAppModeEnum):
            if self.mode_enum is RunAppModeEnum.TEST:
                self._run_test()
            else:
                self._run_app()
        elif isinstance(self.mode_enum, HelperAppModeEnum):
            if self.mode_enum is HelperAppModeEnum.SHELL:
                self._run_shell()
            elif self.mode_enum is HelperAppModeEnum.CMD: 
                # TODO: Custom cmds after assembler build operations.
                raise NotImplementedError
            elif self.mode_enum is HelperAppModeEnum.DEPLOY:
                # TODO: Implement deploy operation.
                raise NotImplementedError
            else:
                raise TypeError
        elif isinstance(self.mode_enum, DatabaseAppModeEnum):
            with self.app.app_context():
                if self.mode_enum is DatabaseAppModeEnum.INIT:
                    self.database.init_migration()
                elif self.mode_enum is DatabaseAppModeEnum.MIGRATE:
                    self.database.migrate_migration()
                elif self.mode_enum is DatabaseAppModeEnum.UPGRADE:
                    self.database.upgrade_migration()
                else:
                    raise TypeError
        else:
            raise TypeError

    def _run_shell(self):
        """Invoke app interactive shell."""
        self.app.run_shell()

    def _run_app(self):
        self.app.run()

    def _run_test(self):
        log.info('Run tests')
        pytest.main(self.mode_args)
        

    def _build_log(self) -> None:
        """Call chain to build log."""
        # Try to find log config cell and build log class from it
        if self.config_classes:
            try:
                log_config_class = Config.find_by_name("log", self.config_classes)
            except ValueError:
                log_config = None
            else:
                # Parse config mapping from cell and append extra configs,
                # if they are given
                app_mode_enum: RunAppModeEnum
                if type(self.mode_enum) is RunAppModeEnum:
                    app_mode_enum = RunAppModeEnum(self.mode_enum.value) 
                else:
                    # Assign dev app mode for all other app modes
                    app_mode_enum = RunAppModeEnum.DEV
                log_config = log_config_class.parse(
                    app_mode_enum=app_mode_enum,
                    root_dir=self.root_dir, 
                    update_with=self.extra_configs_by_name.get("log", None)
                )

            # Init logger
            log_kwargs = log.DEFAULT_LOG_PARAMS

            if log_config:
                for k, v in log_config.items():
                    log_kwargs[k] = v

            log.configure(**log_kwargs)

    def _build_custom_socks(self) -> None:
        if self.sock_classes and self.socket_enabled:
            for sock_class in self.sock_classes:
                socketio: SocketIO = self.socket.get_socketio()

                # Register class for socketio namespace
                # https://flask-socketio.readthedocs.io/en/latest/getting_started.html#class-based-namespaces
                socketio.on_namespace(sock_class(sock_class.NAMESPACE))
                # Also register error handler for the same namespace
                socketio.on_error(sock_class.NAMESPACE)(
                    sock_class.ERROR_HANDLER) 
        elif self.sock_classes and not self.socket_enabled:
            raise AssemblerError(
                'Sock classes are given, but Socket itself is not enabled')
        elif not self.sock_classes and self.socket_enabled:
            log.warning('Socket enabled, but sock classes are not given')

    def _perform_database_postponed_setup(self) -> None:
        """Postponed setup is required, because Database uses Flask app to init
        native SQLAlchemy database inside, so it's possible only after App
        initialization.

        The setup_database requires native flask app to work with.
        """
        self.database.setup(flask_app=self.app.get_native_app())

    def _build_custom_services(self) -> None:
        if self.service_classes:
            for service_class in self.service_classes:
                if self.config_classes:
                    service_config = self._assemble_service_config(
                        name=service_class.get_config_name()) 
                else:
                    service_config = {}

                service: Service = service_class(config=service_config)
                self._custom_services[service_class.get_config_name()] = \
                    service

    def _assemble_service_config(
            self,
            name: str, is_errors_enabled: bool = False) -> dict[str, Any]:
        """Check for service's config in config cells by comparing its given
        name and return it as dict.

        If appropriate config hasn't been found, raise ValueError if
        `is_errors_enabled = True` or return empty dict otherwise.
        """
        try:
            config_class_with_target_name: Config = Config.find_by_name(
                name, self.config_classes)
        except ValueError:
            # If config not found and errors enabled, raise error.
            if is_errors_enabled:
                message = format_message(
                    "Appropriate config for given name {} hasn't been found.",
                    name)
                raise ValueError(message)
            else:
                config: dict[str, Any] = {}
        else:
            if type(config_class_with_target_name) is not Config:
                raise TypeError(format_message(
                    "Type of cell should be Config, but {} received",
                    type(config_class_with_target_name)))
            else:
                app_mode_enum: RunAppModeEnum
                if type(self.mode_enum) is RunAppModeEnum:
                    app_mode_enum = RunAppModeEnum(self.mode_enum.value) 
                else:
                    # Assign dev app mode for all other app modes.
                    app_mode_enum = RunAppModeEnum.DEV
                config = config_class_with_target_name.parse(
                    root_dir=self.root_dir, 
                    update_with=self.extra_configs_by_name.get(name, None),
                    app_mode_enum=app_mode_enum)
    
        # Each builtin service should receive essential fields for their
        # configs, such as root_dir, because they cannot import Assembler
        # due to circular import issue and get this fields by themselves
        if 'root_dir' in config:
            raise ValueError(
                'Service config shouldn\'t contain key `root_dir`')
        config["root_dir"] = self.root_dir

        return config

    def _fetch_yaml_project_version(self) -> str:
        """Fetch project version from info.yaml from the root path and return it. 

        TODO:
            Replace this logic with senseful one. Better use configuration's
            version? Or __init__.py or main.py specified.
        """
        info_data = load_yaml(os.path.join(self.root_dir, "./info.yaml"))
        project_version = info_data["version"]
        return project_version

    def _build_custom_views(self) -> None:
        """Build all views by registering them to app."""
        if self.view_classes:
            for view_class in self.view_classes:
                self.app.register_view(view_class)

    def _build_custom_shell_processors(self) -> None:
        if self.shell_processors:
            self.app.register_shell_processor(*self.shell_processors)

    def _build_custom_cli_cmds(self) -> None:
        # TEMP:
        #   In development
        #
        # if self.cli_cmds:
        #     self.app.register_cli_cmd(*self.cli_cmds)
        pass

    def _build_error_handler(self) -> None:
        ErrorHandler(
            app=self.app,
            default_error_handler=self.default_error_handler,
            default_builtin_error_handler=self.default_builtin_error_hanlder)



import re
import os
import json
from copy import copy
from dataclasses import dataclass
from typing import Any, TypeVar, Sequence

from warepy import load_yaml, get_enum_values

from staze.core.app.app_mode_enum import RunAppModeEnum
from ..assembler.config_extension_enum import ConfigExtensionEnum
from staze.core.model.model import Model
from staze.core.log import log


class Config(Model):
    """Config config which can be used to load configs to appropriate instance's
    configuration by name."""
    name: str
    source_by_app_mode: dict[RunAppModeEnum, str]

    @staticmethod
    def find_by_name(name: str, configs: list['Config']) -> 'Config':
        """Traverse through given list of configs and return first one with
        specified name.
        
        Raise:
            ValueError: 
                No config with given name found.
        """
        for config in configs:
            if config.name == name:
                return config

        raise ValueError(
            "No config found with given name: {}", name)

    def parse(
            self, app_mode_enum: RunAppModeEnum, root_dir: str,
            update_with: dict[str, Any] | None = None,
            convert_keys_to_lower: bool = True) -> dict[str, Any]:
        """Parse config config and return configuration dictionary.

        Args:
            app_mode_enum:
                App mode to run appropriate config.
            root_dir:
                Path to join config config source with.
            update_with (optional):
                Dictionary to update config config mapping with.
                Defaults to None.
            convert_keys_to_lower (optional):
                If true, all keys from origin mapping and mapping from
                `update_with` will be converted to upper case.
        
        Raise:
            ValueError:
                If given config config's source has unrecognized extension.
        """
        res_config: dict[str, Any] = {}

        config_by_mode: dict[RunAppModeEnum, dict] = self._load_config_by_mode()
        res_config = self._update_config_for_mode(config_by_mode, app_mode_enum)

        if res_config:
            self._parse_string_config_values(res_config, root_dir)
        else:
            res_config = {}

        # Update given config with extra dictionary if this dictionary given
        # and not empty.
        if update_with:
            res_config.update(update_with)

        if convert_keys_to_lower:
            temp_config = {}
            for k, v in res_config.items():
                temp_config[k.lower()] = v
            res_config = temp_config

        return res_config

    def _update_config_for_mode(
            self,
            config_by_mode: dict[RunAppModeEnum, dict],
            app_mode_enum: RunAppModeEnum) -> dict:
        """Take config maps for each mode and return result config updated for
        current mode.
        
        E.g., given mode is TEST, so final config will be PROD config updated
        by DEV config and then updated by TEST config (so test keys will
        take priority).
        """ 
        prod_config = copy(config_by_mode[RunAppModeEnum.PROD])

        if app_mode_enum is RunAppModeEnum.TEST:
            dev_config = copy(config_by_mode[RunAppModeEnum.DEV])
            test_config = copy(config_by_mode[RunAppModeEnum.TEST])
            dev_config.update(test_config)
            prod_config.update(dev_config)
        elif app_mode_enum is RunAppModeEnum.DEV:
            dev_config = copy(config_by_mode[RunAppModeEnum.DEV])
            prod_config.update(dev_config)
        else:
            # Prod mode, do nothing extra
            pass
        return prod_config
    
    def _load_config_by_mode(self) -> dict[RunAppModeEnum, dict]:
        config_by_mode: dict[RunAppModeEnum, dict] = {}
        for app_mode_enum in RunAppModeEnum:
            try:
                source = self.source_by_app_mode[app_mode_enum]
            except KeyError:
                # No source for such mode
                config_by_mode[app_mode_enum] = {}
                continue
            source_extension = source[source.rfind(".")+1:]
            config_by_mode[app_mode_enum] = self._load_config_from_file(
                source_extension_enum=ConfigExtensionEnum(source_extension),
                source=source)
        return config_by_mode
    
    def _load_config_from_file(
            self,
            source_extension_enum: ConfigExtensionEnum, source: str) -> dict:
        # Fetch extension and load config from file.
        match source_extension_enum:
            case ConfigExtensionEnum.JSON:
                with open(source, "r") as config_file:
                    config = json.load(config_file)
            case ConfigExtensionEnum.YAML:
                config = load_yaml(source)
                if config is None:
                    raise NotImplementedError(self.name)
            case _:
                raise ValueError("Unrecognized config config source's extension")
        return config
    
    def _parse_string_config_values(
            self, config: dict[str, Any], root_dir: str) -> None:
        for k, v in config.items():
            if type(v) == str:
                # Find environs to be requested
                # Exclude escaped curly brace like `\{not_environ}`
                # Note that environs matching \w+ pattern only supported
                # 
                # Negative look behind used:
                # https://stackoverflow.com/a/3926546
                envs: list[str] = re.findall(r"(?<!\\)\{\w+\}", v)
                if envs:
                    for env in [
                                x.replace("{", "").replace("}", "")
                                for x in envs
                            ]:
                        env = env.strip()
                        real_env_value = os.getenv(env.strip())
                        if real_env_value is None:
                            raise ValueError(
                                f"Environ {env} specified in field"
                                f" {self.name}.{k} was not found")
                        else:
                            v = v.replace("{" + f"{env}" + "}", real_env_value)

                # Look for escaped curly braces and normalize them.
                v = v.replace(r"\{", "{").replace(r"\}", "}")

                # Find paths required to be joined to the root path
                if v[0] == "." and v[1] == "/":
                    config[k] = os.path.join(root_dir, v[2:])
                else:
                    config[k] = v

    @staticmethod
    def find_config_files(
            config_path: str) -> dict[str, dict[RunAppModeEnum, str]]:
        """Accept path to config dir and return dict describing all paths to
        configs for all app modes per service name.
        
        Return example:
        ```python
        {
            "custom_service_name": {
                AppModeEnum.PROD: "./some/path",
                AppModeEnum.DEV: "./some/path",
                AppModeEnum.TEST: "./some/path"
            }
        }
        ```
        """
        source_map_by_name = {} 
        for filename in os.listdir(config_path):
            # Pick files only under config_dir.
            if os.path.isfile(os.path.join(config_path, filename)):
                parts = filename.split(".")
                file_path = os.path.join(config_path, filename)

                if len(parts) == 1:
                    # Skip files without extension.
                    continue
                # Check if file has supported extension.
                elif len(parts) == 2:
                    # Config name shouldn't contain dots and thus we can grab
                    # it right here.
                    config_name = parts[0]
                    if parts[1] in get_enum_values(ConfigExtensionEnum):
                        # Add file without app mode automatically to
                        # `prod`.
                        if config_name not in source_map_by_name:
                            source_map_by_name[config_name] = dict()
                        source_map_by_name[config_name][
                            RunAppModeEnum.PROD] = file_path
                    else:
                        # Skip files with unsupported extension.
                        continue
                elif len(parts) == 3:
                    # Config name shouldn't contain dots and thus we can grab
                    # it right here.
                    config_name = parts[0]
                    if parts[1] in get_enum_values(RunAppModeEnum) \
                            and parts[2] in get_enum_values(
                                ConfigExtensionEnum):
                        # File has both normal extension and defined
                        # app mode.
                        if config_name not in source_map_by_name:
                            source_map_by_name[config_name] = dict()
                        source_map_by_name[config_name][
                            RunAppModeEnum(parts[1])] = file_path
                    else:
                        # Unrecognized app mode or extension,
                        # maybe raise warning?
                        continue
                else:
                    # Skip files with names containing dots, e.g.
                    # "dummy.new.prod.yaml".
                    continue
        return source_map_by_name



from staze.core.app.app_mode_enum import AppModeEnumUnion
from staze.core.model.model import Model


class CliInput(Model):
    mode_enum: AppModeEnumUnion
    mode_args: list[str]

    # Staze's mode args
    host: str = '127.0.0.1'
    port: int = 5000



from dataclasses import dataclass
from pytest import fixture
from staze.core.error.error import Error
from staze.core.test.mock import Mock
from staze.core.log import log
from staze.core import parsing


class ChildError(Error):
    DEFAULT_MESSAGE = 'Something bad happened in child error!'
    DEFAULT_STATUS_CODE = 405


class ErrorMock(Mock):
    message: str
    status_code: int


@fixture
def error_mock() -> ErrorMock:
    return ErrorMock(message='Something bad happened!', status_code=400)


@fixture
def error() -> Error:
    return Error()


@fixture
def error_from_mock(error_mock: ErrorMock) -> Error:
    return Error(error_mock.message, error_mock.status_code)


@fixture
def child_error() -> ChildError:
    return ChildError()


@fixture
def child_error_from_mock(error_mock: ErrorMock) -> ChildError:
    return ChildError(error_mock.message, error_mock.status_code)


class TestError():
    def test_expose(self, error: Error):
        exposed: dict = error.expose()

        exposed_error: dict = parsing.parse_key('error', exposed, dict)

        assert exposed_error['name'] == 'Error'
        assert exposed_error['message'] == Error.DEFAULT_MESSAGE
        assert exposed_error['status_code'] == Error.DEFAULT_STATUS_CODE


    def test_expose_from_mock(
            self, error_from_mock: Error, error_mock: ErrorMock):
        exposed: dict = error_from_mock.expose()

        exposed_error: dict = parsing.parse_key('error', exposed, dict)

        assert exposed_error['name'] == 'Error'
        assert exposed_error['message'] == error_mock.message
        assert exposed_error['status_code'] == error_mock.status_code


class TestChildError():
    def test_expose(self, child_error: ChildError):
        exposed: dict = child_error.expose()

        exposed_error: dict = parsing.parse_key('error', exposed, dict)
        log.debug(exposed)

        assert exposed_error['name'] == 'ChildError'
        assert exposed_error['message'] == ChildError.DEFAULT_MESSAGE
        assert exposed_error['status_code'] == ChildError.DEFAULT_STATUS_CODE


    def test_expose_from_mock(
            self, child_error_from_mock: Error, error_mock: ErrorMock):
        exposed: dict = child_error_from_mock.expose()

        exposed_error: dict = parsing.parse_key('error', exposed, dict)

        assert exposed_error['name'] == 'ChildError'
        assert exposed_error['message'] == error_mock.message
        assert exposed_error['status_code'] == error_mock.status_code



from pytest import fixture
from staze.core.model.model import Model
from staze.core.test.mock import Mock


class CustomModel(Model):
    name: str
    age: int
    tags: set[str]


class CustomModelMock(Mock):
    name: str
    age: int
    tags: set[str]


@fixture
def custom_model_mock() -> CustomModelMock:
    return CustomModelMock(name='Test', age=12, tags={'hero', 'reckless'})


@fixture
def custom_model(custom_model_mock: CustomModelMock) -> CustomModel:
    return CustomModel(
        name=custom_model_mock.name,
        age=custom_model_mock.age,
        tags=custom_model_mock.tags)


class TestModel():
    def test_custom(
            self, custom_model: CustomModel,
            custom_model_mock: CustomModelMock):
        expected_json: dict = {
            'custom': {
                'name': custom_model_mock.name,
                'age': custom_model_mock.age,
                'tags': custom_model_mock.tags
            }
        }

        assert custom_model.name == custom_model_mock.name
        assert custom_model.age == custom_model_mock.age
        assert custom_model.tags == custom_model_mock.tags

        assert custom_model.formatted_name == 'custom'
        assert custom_model.get_api_dict() == expected_json

        json_custom_name: dict = custom_model.get_api_dict('testname')
        assert 'testname' in json_custom_name



from setuptools import setup, find_packages

from staze import __version__ as version


with open("requirements.txt", "r") as file:
    install_requires = [x.strip() for x in file.readlines()]

setup(
    name="staze",
    version=version,
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    description="Upgraded layer over Flask.",
    author="Alexander Ryzhov",
    author_email="thed4rkof@gmail.com",
    url="https://github.com/ryzhovalex/staze",
    keywords=["flask+", "web-framework", "flask", "flask-template"],
    entry_points={
        "console_scripts": [
            "staze = staze.core.cli.cli:main",
        ],
    },
    install_requires=install_requires,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",

        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",

        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.10",
    ],
)



from __future__ import annotations
import re
from functools import wraps
from typing import Callable, Any, TypeVar

from warepy import format_message, snakefy
from staze.core.database.orm_not_found_error import OrmNotFoundError
from staze.core.model.model import Model
from staze.core.log import log
from flask import Flask
import flask_migrate
from flask_sqlalchemy import SQLAlchemy
from flask_sqlalchemy import Model as BaseOrm
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declared_attr

from staze.core.service.service import Service
from .database_type_enum import DatabaseTypeEnum


AnyOrm = TypeVar('AnyOrm', bound='Database.Orm')


# TODO: Fix type hinting for decorated functions under this decorator.
def migration_implemented(func: Callable):
    @wraps(func)
    def inner(self_instance, *args, **kwargs):
        if type(self_instance) is not Database:
            raise TypeError(
                "Decorator migration_implemented cannot be"
                f" applied to type {type(self_instance)}")
        elif not self_instance.migration:
            error_message = "Migrate object hasn't been implemented yet"
            raise AttributeError(error_message)
        else:
            return func(self_instance, *args, **kwargs)
    return inner


class Orm(BaseOrm):
    """Base orm model responsible of holding database model's data and at least
    it's basic CRUD operations.

    Contains create(), get_first(), get_all() and del_first() methods as
    Create, Retrieve and Delete representatives.
    Update representatives are defined individually at each subclass
    (e.g. `set_something()`), and by default accessed via basic model
    alteration, e.g. `MyModel.name = 'Another name'`.
    """
    # As superclass sqlalchemy used instead of `orm` class to avoid reference
    # errors
    # https://flask-sqlalchemy.palletsprojects.com/en/2.x/customizing/

    # Name to exclude on tablename forming
    _BASE_NAME: str = 'Orm'

    id = sa.Column(sa.Integer, primary_key=True)
    type = sa.Column(sa.String(250))

    @declared_attr
    def __tablename__(cls) -> str:
        cls_name: str = cls.__name__  # type: ignore
        return snakefy(cls_name)

    @declared_attr
    def __identity__(cls) -> str:
        return snakefy(cls.__name__.replace(cls._BASE_NAME, ''))  # type: ignore

    @declared_attr
    def __mapper_args__(cls) -> dict[str, Any]:
        args: dict[str, Any] = {}
        args.update({
            'polymorphic_on': 'type',
            'polymorphic_identity': cls.__identity__
        })
        return args

    @classmethod
    def create(cls: AnyOrm, **kwargs) -> AnyOrm:
        """Create model and return it.
        
        Accepts all given kwargs and thus is recommended to be redefined at
        subclasses.

        Model creation intended to be done only through this method.
        """
        model = cls(**kwargs)
        Database.instance().push(model)
        return model

    @classmethod
    def get_first(
            cls,
            order_by: object | list[object] | None = None,
            **kwargs) -> Database.Orm:
        """Filter first ORM orm model by given kwargs and return it.
        
        Raise:
            ValueError:
                No such ORM model in database matched given kwargs
        """
        query: Any = cls.query.filter_by(**kwargs)  # type: ignore

        if order_by is not None:
            query = cls._order_query(query, order_by)

        model: Database.Orm = query.first()

        if not model:
            raise OrmNotFoundError(orm_name=cls.__name__, **kwargs)
        else:
            return model

    @classmethod
    def get_all(
            cls,
            order_by: object | list[object] | None = None,
            limit: int | None = None,
            **kwargs) -> list[Database.Orm]:
        """Filter all ORM orm models by given kwargs and return them.

        Return:
            List of found models.
            If no models found, empty list is returned.
        """
        query: Any = cls.query.filter_by(**kwargs)  # type: ignore

        if order_by is not None:
            query = cls._order_query(query, order_by)

            if limit:
                query = query.limit(limit)
        elif limit:
            query = query.limit(limit)

        models: list[Database.Orm] = query.all()

        if type(models) is not list:
            raise OrmNotFoundError(model_name=cls.__name__, **kwargs)
        else:
            # Return models even if it's empty list
            return models

    @classmethod
    def del_first(
            cls,
            order_by: object | list[object] | None = None,
            **kwargs) -> None:
        """Delete first accessed by `get_first()` method model."""
        database: Database = Database.instance()
        model: Database.Orm = cls.get_first(order_by=order_by, **kwargs)

        database.native_database.session.delete(model)
        database.commit()

    @staticmethod
    def _order_query(query: Any, order_by: object | list[object]) -> object:
        if type(order_by) is list:
            return query.order_by(*order_by)
        else:
            return query.order_by(order_by)

    @property
    def model(cls) -> Model:
        """Transform self to a Model instance.
        
        Model here should be defined explicitly at subclass of superclass
        Model.

        Example:
            UserOrm has to define User(Model) subclass and redefine this
            method to construct User(Model) class with values from the Orm
            it requires.
            It's often useful for API calls to Orms.

        Returns:
            Model:
                Model subclass contained required to expose Orm's
                properties.
        """
        raise NotImplementedError(
            'Should be re-implemented for Orm-specific Model subclass')


class Database(Service):
    """Operates over database processes."""
    # Helper references for shorter writing at ORMs.
    # Ignore lines added for a workaround to fix issue:
    # https://github.com/microsoft/pylance-release/issues/187
    native_database = SQLAlchemy(model_class=Orm)
    Orm: Any = native_database.Model 
    column = native_database.Column
    integer = native_database.Integer
    string = native_database.String
    text = native_database.Text
    float = native_database.Float
    boolean = native_database.Boolean
    foreign_key = native_database.ForeignKey
    table = native_database.Table
    check_constraint = native_database.CheckConstraint
    relationship = native_database.relationship
    backref = native_database.backref
    pickle = native_database.PickleType
    binary = native_database.LargeBinary
    datetime = native_database.DateTime

    def __init__(self, config: dict) -> None:
        super().__init__(config)
        self.DEFAULT_URI = f"sqlite:///{self.config['root_dir']}/sqlite3.database"

        self.native_database = Database.native_database
        # For now service config propagated to Database domain
        self._assign_uri_from_config(config)

    def _assign_uri_from_config(self, config: dict) -> None:
        raw_uri = config.get("uri", None)  # type: str

        if not raw_uri:
            log.info(f"URI for database is not specified, using default")
            raw_uri = self.DEFAULT_URI
        else:
            # Case 1: SQLite Database
            # Developer can give relative path to the Database
            # (it will be absolutized at Config.parse()),
            # by setting sqlite Database extension to `.database`, e.g.:
            #   `./instance/sqlite3.database`
            # or by setting full absolute path with protocol, e.g.:
            #   `sqlite:////home/user/project/instance/sqlite3.database`
            if raw_uri.rfind(".database") != -1 or "sqlite:///" in raw_uri:
                if "sqlite:///" not in raw_uri: 
                    # Set absolute path to database
                    # Ref: https://stackoverflow.com/a/44687471/14748231
                    self.uri = "sqlite:///" + raw_uri
                else:
                    self.uri = raw_uri
                self.type_enum = DatabaseTypeEnum.SQLITE
            # Case 2: PostgreSQL Database
            elif re.match(r"postgresql(\+\w+)?://", raw_uri):
                # No need to calculate path since psql uri should be given in
                # full form
                self.uri = raw_uri
                self.type_enum = DatabaseTypeEnum.PSQL
            else:
                raise ValueError(
                    "Unrecognized or yet unsupported type of Database uri:"
                    f" {raw_uri}")
            
            # WARNING:
            #   Never print full Database uri to config, since it may
            #   contain user's password (as in case of psql)
            log.info(f"Set database type: {self.type_enum.value}")

    @migration_implemented
    def init_migration(
            self,
            directory: str = "migrations",
            multidatabase: bool = False) -> None:
        """Initializes migration support for the application."""
        flask_migrate.init(directory=directory, multidb=multidatabase)

    @migration_implemented
    def migrate_migration(
        self,
        directory: str = "migrations", 
        message = None, 
        sql = False, 
        head: str = "head", 
        splice = False, 
        branch_label = None, 
        version_path = None, 
        rev_id = None
    ) -> None:
        flask_migrate.migrate(
            directory=directory,
            message=message,
            sql=sql,
            head=head,
            splice=splice,
            branch_label=branch_label,
            version_path=version_path,
            rev_id=rev_id
        )

    @migration_implemented
    def upgrade_migration(
        self,
        directory: str = "migrations",
        revision: str = "head",
        sql = False,
        tag = None
    ) -> None:
        flask_migrate.upgrade(
            directory=directory,
            revision=revision,
            sql=sql,
            tag=tag
        )

    def setup(self, flask_app: Flask) -> None:
        """Setup Database and migration object with given Flask app."""
        flask_app.config["SQLALCHEMY_DATABASE_URI"] = self.uri
        flask_app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
        self.native_database.init_app(flask_app)

        # render_as_batch kwarg required only for sqlite3 databases to avoid
        # ALTER TABLE issue on migrations
        # https://blog.miguelgrinberg.com/post/fixing-alter-table-errors-with-flask-migrate-and-sqlite
        if self.type_enum is DatabaseTypeEnum.SQLITE:
            is_sqlite_database = True
        else:
            is_sqlite_database = False
        self.migration = flask_migrate.Migrate(flask_app, self.native_database, render_as_batch=is_sqlite_database)

    def get_native_database(self) -> SQLAlchemy:
        return self.native_database

    @migration_implemented
    def get_migration(self) -> flask_migrate.Migrate:
        """Return migration object.
        
        Raise:
            AttributeError: If Migrate object hasn't implemented yet.
        """
        return self.migration

    @migration_implemented
    def create_all(self) -> None:
        self.native_database.create_all()

    @migration_implemented
    def drop_all(self):
        "Drop all tables."
        self.native_database.drop_all()

    @migration_implemented
    def _add(self, entity):
        """Place an object in the session."""
        # TODO: Add functionality to accept multiple entities as *args.
        self.native_database.session.add(entity)

    def push(self, entity):
        """Add entity to session and immediately commit the session."""
        self._add(entity)
        self.commit()

    @migration_implemented
    def commit(self):
        """Commit current transaction."""
        self.native_database.session.commit()

    @migration_implemented
    def rollback(self):
        self.native_database.session.rollback()

    @migration_implemented
    def remove(self):
        self.native_database.session.remove()



from typing import Callable


class Error(Exception):
    """Main error class to be inherited in all app's custom errors.
    
    Args:
        message (str, optional):
            Message of the error. Defaults to cls.DEFAULT_MESSAGE
        status_code (int, optional):
            Status code of the error. Defaults to cls.DEFAULT_STATUS_CODE

    Attributes:
        message (str):
            Message of the error. Defaults to cls.DEFAULT_MESSAGE
        status_code (int):
            Status code of the error. Defaults to cls.DEFAULT_STATUS_CODE

    Raises:
        TypeError:
            If message or status code types are incorrect.
    """
    DEFAULT_MESSAGE = ''
    DEFAULT_STATUS_CODE = 400
    HANDLER_FUNCTION: Callable | None = None
    SHOULD_BE_LOGGED: bool = True

    def __init__(
            self,
            message: str | None = None,
            status_code: int | None = None) -> None:
        super().__init__(message)

        self.message: str
        self.status_code: int

        if message is None:
            self.message = self.DEFAULT_MESSAGE
        elif type(message) is str:
            self.message = message
        else:
            raise TypeError(
                'Error first argument (message) should be str,'
                f' got {type(message)} instead')

        if status_code is None:
            self.status_code = self.DEFAULT_STATUS_CODE
        elif type(status_code) is int:
            self.status_code: int = status_code
        else:
            raise TypeError(
                'Error second argument (status_code) should be int,'
                f' got {type(status_code)} instead')

    def expose(self) -> dict:
        """Expose error to structured dict for API.

        Returns:
            dict:
                Dictionary with structure:
        ```
        {
            'error': {
                'name': error_class_name,
                'message': error_message,
                'status_code': error_status_code
            }
        }
        ```
        """
        return {"error": {
            "name": self.__class__.__name__,
            "message": self.message,
            "status_code": self.status_code
        }}

    def __str__(self) -> str:
        return self.message



from staze.core.error.error import Error
from typing import Callable

from staze.core.app.app import App
from staze.core.log import log, raise_error_to_log
from staze.core.error.error import Error
from staze.core.service.service import Service
from warepy import Singleton


class ErrorHandler(Singleton):
    def __init__(
            self,
            app: App,
            default_error_handler: Callable | None = None,
            default_builtin_error_handler: Callable | None = None) -> None:
        self.default_error_handler: Callable
        self.default_builtin_error_handler: Callable

        if isinstance(default_error_handler, Callable):
            self.default_error_handler = default_error_handler
        else:
            self.default_error_handler = self._handle_error_default

        if isinstance(default_builtin_error_handler, Callable):
            self.default_builtin_error_handler = default_builtin_error_handler
        else:
            self.default_builtin_error_handler = \
                self._handle_builtin_error_default

        # Register main handler function
        app.register_error(Exception, self.handle_error)
        
    def handle_error(self, err: Exception):
        if isinstance(err, Error):
            if err.SHOULD_BE_LOGGED:
                raise_error_to_log(err)
            if err.HANDLER_FUNCTION:
                return err.HANDLER_FUNCTION(err)
            else:
                # Use default
                return self.default_error_handler(err)
        else:
            # Log builtin exceptions in any case
            raise_error_to_log(err)
            return self.default_builtin_error_handler(err)

    def _handle_error_default(self, err: Error):
        return err.expose(), err.status_code

    def _handle_builtin_error_default(self, err: Exception):
        return self._handle_error_default(
            Error('; '.join(err.args), 400))



from warepy import Singleton
from staze.core.log import log


class Service(Singleton):
    """Service superclass inherits Singleton metaclass.

    Layer between Domain Objects (Domains) and UI layer
    (Controllers and Views).
    Performs various operations e.g. to prepare data (like json formatting)
    coming to Domains and from them.
    """
    # Name of config to be fetched for this service, if None, builded from
    # class name with excluded BASE_NAME, e.g.: 'UserService' -> 'user',
    # and thus e.g. `user.yaml` will be searched in config folder
    CONFIG_NAME: str | None = None
    BASE_NAME: str = 'Service'

    def __init__(self, config: dict) -> None:
        self.config = config

    @classmethod
    def get_config_name(cls) -> str:
        if cls.CONFIG_NAME:
            return cls.CONFIG_NAME
        else:
            return cls.__name__.replace(cls.BASE_NAME, '').lower()
